/* Le but de ce programme est de demontrer les possibilitées des bibliotheques vkey et wavtrigger.
    Nous envisageons ici un processus d'acquisition au clavier d'une valeur numerique
    puis de sa lecture via une carte son en respectant les bizareries
    de la langue française sur la lecture des nombres.
    nous nous limiterons à des nombres à virgule inférieurs en valeur absolue à
    999 999 999 999,99
    neuf-cent-quatre-vingt-dix-neuf milliard neuf-cent-quatre-vingt-dix-neuf millions
    neuf-cent-quatre-vingt-dix-neuf mille neuf-cent-quatre-vingt-dix-neuf
    virgule quatre-vingt-dix-neuf
    le clavier envisagé n'ayant que 12 touches, les dix digits+le symbole decimal+le return
    ne permettent pas de prendre les nombres <0.
    Néanmoins en prenant un clavier adapté (16 touches) ou un bouton de validation externe,
    il est possible sans modification du coeur du programme de prendre les nombres negatifs.
*/

// ******************************************
// Les options suivant le matériel connecté

#define carteson // ligne à decommenter pour l'application sur arduino avec carte son
#define ecran // ligne à decommenter pour l'application sur arduino avec ecran de control serie
#define control // ligne à decommenter pour l'application sur arduino avec sortie sur ecran PC

// *********************
// Les bibliotheques

#include <math.h>
#include <vkey.h>

// *********************

#define brocheclavier A6 // connecté à la broche A6
// Un bouton poussoir
#define Bp 15
// Une led
#define Led 14

#define vitessevisu 9600 // vitesse de communication avec le moniteur

// *********************

#define nbavantvirgule 12 // nombres codés avec 12 chiffres maxi avant la virgule (999 milliards)
#define nbdecimale 2 // et deux decimales

// ********************
// Les touches du clavier et leur correspondances

#define valide 12
#define pointdecimal 11
#define touchezero 10
#define toucheun 1
#define touchedeux 2
#define touchetrois 3
#define touchequatre 4
#define touchecinq 5
#define touchesix 6
#define touchesept 7
#define touchehuit 8
#define toucheneuf 9
#define toucherien 0

// Les definitions des numeros des fichiers son en fonction de leur contenu
#define _cent 100
#define _mille 101
#define _million 102
#define _milliard 103
#define _et 104
#define _euro 105
#define _centime 106
#define _moins 107

// *******************************************************************************
// Les objets utils

Vkey key = Vkey(brocheclavier); // création de l'objet key connecté sur broche A0

#ifdef carteson
// definition de la carte son
#include <wavTrigger.h>
wavTrigger  carteWave = wavTrigger(&Serial1, 9600); // carte son branchée sur Serial1
#endif
// de carteson

#ifdef ecran
// definition de l'ecran de contrôle à 9600 bds sur la broche 3
#include <SerialDisplay.h>
SerialDisplay ecran_serie = SerialDisplay(16);
#endif

// *************************************************************
// prend un nombre entier au clavier
// contenant au maxi nbdigitmax symboles
// si prisezero est vrai, on compte les premiers zero
// si non, on ne prend en compte que les chiffres significatifs

long long  prendEntier(int nbdigitsmax, boolean prisezero)
{
  static int table[10] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 0};
  long long  N = 0;
  int nbDigits = 0;
  int d = 0;
  do
  { d = key.getDigitFmt(); // on va chercher la touche enfoncée
    if ((d > toucherien) & (d <= touchezero)) // il sagit d'un symbole numerique
    {
      N = N * 10 + table[d - 1]; // on decale et ajoute la bonne valeur
      if (!((table[d - 1] == 0) && (nbDigits == 0) && prisezero) ) // on ne prends pas les zero en debut de nombre sauf si demande
      {
        nbDigits++; // nb de digits pris en compte
#ifdef ecran
        Serial.print(table[d - 1]); // visualisation
#endif
      }
    }
  }
  while ((nbDigits < nbdigitsmax) & (d <= touchezero));
#ifdef ecran
  if ((N == 0) && (nbDigits == 0)) // cas extreme on a tappé sur '.' ou enter
  {
    Serial.print('0');
  }
#endif
  return N;
}

// *********************************************************
// prend la valeur entière et la valeur décimale d'un nombre
// avec nent digits au maxi avant la virgule
// et ndec digits maxi apres la virgule
// gère les cas <enter>, .xy<enter> , 0.<enter>, .<enter> ...

void prendNombre(long long &ent, long long &dec, byte nent, byte ndec)
{
  Serial.print('[');
  ent = prendEntier(nent, true); // prise de la valeur entiere
  int der = key.getDerniereTouche(); // pour savoir comment on continu
  switch (der)
  {
    case valide : // enter enfoncé
      {
        dec = 0;
        break;
      }
    default: // derniere touche 1..10 et nombre nent atteint ou point decimal enfoncé
      {
        Serial.print('.');
        dec = prendEntier(ndec, false);
        break;
      }
  }

  Serial.println(']');
}

// *********************************************************

#ifdef ecran
// procedure qui visualise les phonemes sur l'ecran lcd
void show_ecran(int n)
{
  ecran_serie.Posxy(1, 2);
  ecran_serie.print("F: "); ecran_serie.Clreoln();
  ecran_serie.Posxy(3, 2);
  switch (n)
  {
    case _et :
      ecran_serie.print("et");
      break;

    case _cent :
      ecran_serie.print("cent");
      break;

    case _mille :
      ecran_serie.print("mille");
      break;

    case _million :
      ecran_serie.print("millions" );
      break;

    case _milliard :
      ecran_serie.print("milliards");
      break;

    case _euro :
      ecran_serie.print("euro");
      break;

    case _centime :
      ecran_serie.print("centime");
      break;

    case _moins :
      ecran_serie.print("moins");
      break;

    default :

      ecran_serie.print(n);
  }
}
#endif // de ecran

// *********************************************************

#ifdef control
// procedure qui visualise les phonemes sur l'ecran du PC
void show_serial(int n)
{
  Serial.print("fichier : ");
  switch (n)
  {
    case _et :
      Serial.println(" et ");
      break;

    case _cent :
      Serial.println(" cent ");
      break;

    case _mille :
      Serial.println(" mille ");
      break;

    case _million :
      Serial.println(" millions ");
      break;

    case _milliard :
      Serial.println(" milliards" );
      break;

    case _euro :
      Serial.println("euro ");
      break;

    case _centime :
      Serial.println("centime ");
      break;

    case _moins :
      Serial.println("moins ");
      break;
    default :

      Serial.println( n);
  }
}
#endif // de control

// *********************************************************

void dit(int n) // on prononce ou enonce les phonèmes/mots
{
  // montre les fichiers qui doivent être lus
  // sur l'ecran serie
#ifdef ecran
  show_ecran(n);
#endif

  // on the serial to PC
#ifdef control
  show_serial(n);
#endif

#ifdef carteson
  // lit le fichier son
  carteWave.trackPlaySolo(n);
  // attend que le fichier soit fini de lire
  // en attendant la réponse de la carte son

  while (carteWave.isTrackPlaying(n)) {
    delay(5);
  };
#endif
}
// *********************************************************

// traitement des nombres de 0 à 16
void zero_16(byte n, bool z) // traitement de 0 à 16
{
  if ( (n != 0) || z)
  {
    dit(n);
  }

}
// *********************************************************

// traitement des nombres de 0 à 19
void zero_19(byte n, bool z) // traitement de 0 à 19
{
  if (n < 17) {
    zero_16(n, z);
  }
  else
  {
    dit(10);
    zero_16(n - 10, false);
  }

}
// *********************************************************

// traitement des nombres de 0 à 99
// la partie entiere de ce nombre peut entrer sur un byte
// 99 < 254
// 99 < max_byte

void zero_99(byte n, bool z) // traitement de 0 à 99 (cent -1)
{ if (n < 20)
  {
    zero_19(n, z);
  }
  else
  {
    int dixaine = n / 10;
    if ( (dixaine == 9) || (dixaine == 7)) // le cas de septante et nonante qui se disent
      // respectivement soixante <reste> ou quatre vingt <reste> ou reste est un nombre entre 0 et 20
    {
      dixaine--;
    };

    if (dixaine == 8) // le cas de octante qui se dit quatre vingt
    {
      dit (4);
      dit (20);
    }
    else
    {
      dit(dixaine * 10);
    }
    int reste = n - dixaine * 10;
    if (reste > 0)
    {
      if ((reste == 1) & (dixaine != 8)) {
        dit(_et); // on dit cinquante et un et non quate-vingt-et-un
      }
      zero_19(reste, false);
    }
  }
}

// *********************************************************

// traitement de 0 à 999 (mille -1)
// la partie entiere de ce nombre peut entrer sur un unsigned int
// 999 < 65535
// 999 < max_unsigned int

void zero_999(unsigned int n, bool z)

{ if (n > 99)
  {
    int  centaines = n / 100;
    if (centaines > 1)
    {
      dit(centaines); // on ne dit pas 1 cent mais cent
    }
    dit(_cent);
    n = n % 100;
    z = false;
  }
  zero_99(n, false);
}

// *********************************************************
// traitement de 0 à 999999 (1millios -1)
// la partie entiere de ce nombre peut entrer sur un unsigned long
// 999999 < 4294967295
// 999999 < max_unsigned_long

void zero_999999(unsigned long n, bool z)
{
  if (n > 999)
  {
    long  millier = n / 1000L;
    if (millier > 1)
    {
      zero_999(millier, false);// on ne dit pas 1 mille mais mille !!
    }
    dit (_mille);

    n = n % 1000LL;
    z = false;
  }
  zero_999(n, z);
}
// *********************************************************
// traitement de 0 à 999999999  (1 milliard -1)
// la partie entiere de ce nombre peut entrer sur un unsigned long
// 999999999 < 4294967295

void zero_999999999(unsigned long long n, bool z)
{
  if (n > 999999LL)
  {
    long  million = n / 1000000LL;
    zero_999(million, false);
    dit (_million); // ici on dit 1 million
    n = n % 1000000LL;
    z = false;
  }
  zero_999999(n, z);
}


// *********************************************************
// traitement de 0 à 999999999999 (1 trillon-1)
// la partie entiere de ce nombre doit rentrer sur un format > au format long
// soit long long
// 4294967295 < 999999999999 < 18446744073709551615
// max_unsigned_long < 999999999999 < max_unsigned_long_long

void zero_999999999999(unsigned long long n, bool z)
{
  if (n > 999999999LL)
  {
    long  milliard = n / 1000000000LL;
    zero_999(milliard, false);
    dit (_milliard); // ici on dit 1 milliard
    n = n % 1000000000LL;
    z = false;
  }
  zero_999999999(n, z);
}

// *********************************************************
// enonce un un prix en euros en fonction de la valeur entière
// et de la valeur decimale
void enonce_prix(long long  valent, unsigned long decimale)
{
  if (valent < 0)
  {
    valent = -valent;
    dit(_moins);
  }

  zero_999999999999(valent, true);
  dit(_euro);
  if (decimale > 0)
  {
    dit(_et);
    zero_999(decimale, true);
    dit(_centime);
  }
}

// **************************************
// Le initialisations
void setup()
{
#ifdef ecran
  ecran_serie.begin(9600); // initialisation de l'écran serie
  ecran_serie.Clrscr(); // efface l'écran de controle
#endif // de #ifdef ecran

#ifdef carteson
  carteWave.start(9600);
  delay(1000);
  carteWave.setReporting(true); // active le mode retour information de la carte wav
  delay(1000);
  char v[50];
  carteWave.getVersion(v, 50);
#endif

#ifdef control
  Serial.begin(9600);
#ifdef carteson
  Serial.println(v); // affiche la version du firmware de la carte son
#endif
#endif
  // Les entrées sorties tor
  pinMode(Bp, INPUT_PULLUP);
  pinMode(Led, OUTPUT);
  digitalWrite(Led, false);
}

// **************************************************************
// Procedure principale
// executée en boucle

void loop() // procedure appelée de manière cyclique par main
{

  long long ent; // un entier entre -9223372036854775808 et 9223372036854775807 18 digits decimaux possibles 1 trillon !!
  long long dec;
  long sto;
  prendNombre(ent, dec, nbavantvirgule, nbdecimale ); // prends une valeur réelle au clavier

  // attention les procedures standard print ne prennent pas en compte les long long,
  // il convient donc de biaiser pour les imprimer en tant que deux long
  // l'un représentant les poids forts
  // l'autre les poids faibles
  // en faisant atention de n'imprimer le signe qu'une fois !

#ifdef ecran
  ecran_serie.Posxy(1, 1);
  ecran_serie.print("[");
  if (ent < 0)
  {
    ecran_serie.print('-');
  }
  sto = abs(ent) / 0xFFFFFFFFFF;
  if (abs(sto) > 0) {
    ecran_serie.print(sto);
  }
  sto = abs(ent) % 0xFFFFFFFFFF;
  ecran_serie.print(sto);
  ecran_serie.print(".");
  if (dec < 10)
  {
    ecran_serie.print("0");
  }
  ecran_serie.print(long(dec));
  ecran_serie.println("]");

#endif
#ifdef control
  Serial.println("**************************");
  Serial.print("[");
  if (ent < 0)
  {
    Serial.print('-');
  }
  sto = abs(ent) / 0xFFFFFFFFFF;
  if (abs(sto) > 0)
  {
    Serial.print(sto);
  }
  sto = abs(ent) % 0xFFFFFFFFFF;
  Serial.print(sto);
  Serial.print(".");
  if (dec < 10)
  {
    Serial.print("0");
  }
  Serial.print(long(dec));
  Serial.println("]");
#endif
  enonce_prix(ent, dec);
  digitalWrite(Led, false);
  while (digitalRead(Bp)) {};

}

/* La procedure main cachée
    int main()
    {
    setup();
    while (true) {loop();}
    }

*/



